@@ -29,7 +29,7 @@ from builtins import isinstance
 from typing import List, Iterable
 
 from keycloak.urls_patterns import URL_ADMIN_GROUPS_REALM_ROLES, \
-    URL_ADMIN_GET_GROUPS_REALM_ROLES, URL_ADMIN_REALM_ROLES_ROLE_BY_NAME
+    URL_ADMIN_GET_GROUPS_REALM_ROLES, URL_ADMIN_REALM_ROLES_ROLE_BY_NAME, URL_ADMIN_GROUPS_CLIENT_ROLES
 from .connection import ConnectionManager
 from .exceptions import raise_error_from_response, KeycloakGetError
 from .keycloak_openid import KeycloakOpenID
@@ -42,7 +42,8 @@ from .urls_patterns import URL_ADMIN_SERVER_INFO, URL_ADMIN_CLIENT_AUTHZ_RESOURC
     URL_ADMIN_GROUP_MEMBERS, URL_ADMIN_USER_STORAGE, URL_ADMIN_GROUP_PERMISSIONS, URL_ADMIN_IDPS, \
     URL_ADMIN_USER_CLIENT_ROLES_AVAILABLE, URL_ADMIN_USERS, URL_ADMIN_CLIENT_SCOPES, \
     URL_ADMIN_CLIENT_SCOPES_ADD_MAPPER, URL_ADMIN_CLIENT_SCOPE, URL_ADMIN_CLIENT_SECRETS, \
-    URL_ADMIN_USER_REALM_ROLES, URL_ADMIN_REALM, URL_ADMIN_COMPONENTS, URL_ADMIN_COMPONENT, URL_ADMIN_KEYS 
+    URL_ADMIN_USER_REALM_ROLES, URL_ADMIN_REALM, URL_ADMIN_COMPONENTS, URL_ADMIN_COMPONENT, URL_ADMIN_KEYS, \
+    URL_ADMIN_USER_FEDERATED_IDENTITY, URL_ADMIN_USER_FEDERATED_IDENTITIES
 
 
 class KeycloakAdmin:
@@ -193,7 +194,6 @@ class KeycloakAdmin:
 
         self._auto_refresh_token = value
 
-
     def __fetch_all(self, url, query=None):
         '''Wrapper function to paginate GET requests
 
@@ -236,7 +236,7 @@ class KeycloakAdmin:
 
         data_raw = self.raw_post(URL_ADMIN_REALMS,
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201])
 
     def get_realms(self):
         """
@@ -261,7 +261,7 @@ class KeycloakAdmin:
 
         data_raw = self.raw_post(URL_ADMIN_REALMS,
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201], skip_exists=skip_exists)
 
     def update_realm(self, realm_name, payload):
         """
@@ -279,7 +279,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": realm_name}
         data_raw = self.raw_put(URL_ADMIN_REALM.format(**params_path),
                                 data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_realm(self, realm_name):
         """
@@ -291,7 +291,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": realm_name}
         data_raw = self.raw_delete(URL_ADMIN_REALM.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def get_users(self, query=None):
         """
@@ -339,7 +339,7 @@ class KeycloakAdmin:
 
         data_raw = self.raw_post(URL_ADMIN_USERS.format(**params_path),
                                  data=json.dumps(payload))
-        raise_error_from_response(data_raw, KeycloakGetError, expected_code=201)
+        raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201])
         _last_slash_idx = data_raw.headers['Location'].rindex('/')
         return data_raw.headers['Location'][_last_slash_idx + 1:]
 
@@ -408,7 +408,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": user_id}
         data_raw = self.raw_put(URL_ADMIN_USER.format(**params_path),
                                 data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_user(self, user_id):
         """
@@ -420,7 +420,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": user_id}
         data_raw = self.raw_delete(URL_ADMIN_USER.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def set_user_password(self, user_id, password, temporary=True):
         """
@@ -440,7 +440,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": user_id}
         data_raw = self.raw_put(URL_ADMIN_RESET_PASSWORD.format(**params_path),
                                 data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def consents_user(self, user_id):
         """
@@ -454,6 +454,30 @@ class KeycloakAdmin:
         data_raw = self.raw_get(URL_ADMIN_USER_CONSENTS.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
+    def get_user_social_logins(self, user_id):
+        """
+        Returns a list of federated identities/social logins of which the user has been associated with
+        :param user_id: User id
+        :return: federated identities list
+        """
+        params_path = {"realm-name": self.realm_name, "id": user_id}
+        data_raw = self.raw_get(URL_ADMIN_USER_FEDERATED_IDENTITIES.format(**params_path))
+        return raise_error_from_response(data_raw, KeycloakGetError)
+
+    def add_user_social_login(self, user_id, provider_id, provider_userid, provider_username):
+
+        """
+        Add a federated identity / social login provider to the user
+        :param user_id: User id
+        :param provider_id: Social login provider id
+        :param provider_userid: userid specified by the provider
+        :param provider_username: username specified by the provider
+        :return:
+        """
+        payload = {"identityProvider": provider_id, "userId": provider_userid, "userName": provider_username}
+        params_path = {"realm-name": self.realm_name, "id": user_id, "provider": provider_id}
+        data_raw = self.raw_post(URL_ADMIN_USER_FEDERATED_IDENTITY.format(**params_path), data=json.dumps(payload))
+    
     def send_update_account(self, user_id, payload, client_id=None, lifespan=None, redirect_uri=None):
         """
         Send an update account email to the user. An email contains a
@@ -633,7 +657,7 @@ class KeycloakAdmin:
             data_raw = self.raw_post(URL_ADMIN_GROUP_CHILD.format(**params_path),
                                      data=json.dumps(payload))
 
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201], skip_exists=skip_exists)
 
     def update_group(self, group_id, payload):
         """
@@ -651,7 +675,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": group_id}
         data_raw = self.raw_put(URL_ADMIN_GROUP.format(**params_path),
                                 data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def group_set_permissions(self, group_id, enabled=True):
         """
@@ -678,7 +702,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": user_id, "group-id": group_id}
         data_raw = self.raw_put(URL_ADMIN_USER_GROUP.format(**params_path), data=None)
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def group_user_remove(self, user_id, group_id):
         """
@@ -691,7 +715,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": user_id, "group-id": group_id}
         data_raw = self.raw_delete(URL_ADMIN_USER_GROUP.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_group(self, group_id):
         """
@@ -703,7 +727,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": group_id}
         data_raw = self.raw_delete(URL_ADMIN_GROUP.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def get_clients(self):
         """
@@ -792,7 +816,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name}
         data_raw = self.raw_post(URL_ADMIN_CLIENTS.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201], skip_exists=skip_exists)
 
     def update_client(self, client_id, payload):
         """
@@ -806,7 +830,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_put(URL_ADMIN_CLIENT.format(**params_path),
                                            data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_client(self, client_id):
         """
@@ -821,7 +845,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_delete(URL_ADMIN_CLIENT.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def get_realm_roles(self):
         """
@@ -904,7 +928,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": client_role_id}
         data_raw = self.raw_post(URL_ADMIN_CLIENT_ROLES.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201], skip_exists=skip_exists)
 
     def delete_client_role(self, client_role_id, role_name):
         """
@@ -918,7 +942,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": client_role_id, "role-name": role_name}
         data_raw = self.raw_delete(URL_ADMIN_CLIENT_ROLE.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def assign_client_role(self, user_id, client_id, roles):
         """
@@ -934,7 +958,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": user_id, "client-id": client_id}
         data_raw = self.raw_post(URL_ADMIN_USER_CLIENT_ROLES.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def create_realm_role(self, payload, skip_exists=False):
         """
@@ -948,7 +972,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name}
         data_raw = self.raw_post(URL_ADMIN_REALM_ROLES.format(**params_path),
                                             data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201], skip_exists=skip_exists)
 
     def update_realm_role(self, role_name, payload):
         """
@@ -961,7 +985,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "role-name": role_name}
         data_raw = self.connection.raw_put(URL_ADMIN_REALM_ROLES_ROLE_BY_NAME.format(**params_path),
                                             data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_realm_role(self, role_name):
         """
@@ -973,7 +997,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "role-name": role_name}
         data_raw = self.connection.raw_delete(
             URL_ADMIN_REALM_ROLES_ROLE_BY_NAME.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def assign_realm_roles(self, user_id, client_id, roles):
         """
@@ -989,7 +1013,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": user_id}
         data_raw = self.raw_post(URL_ADMIN_USER_REALM_ROLES.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def assign_group_realm_roles(self, group_id, roles):
         """
@@ -1004,7 +1028,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": group_id}
         data_raw = self.raw_post(URL_ADMIN_GROUPS_REALM_ROLES.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_group_realm_roles(self, group_id, roles):
         """
@@ -1019,7 +1043,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": group_id}
         data_raw = self.raw_delete(URL_ADMIN_GROUPS_REALM_ROLES.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def get_group_realm_roles(self, group_id):
         """
@@ -1032,6 +1056,53 @@ class KeycloakAdmin:
         data_raw = self.raw_get(URL_ADMIN_GET_GROUPS_REALM_ROLES.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
+    def assign_group_client_roles(self, group_id, client_id, roles):
+        """
+        Assign client roles to a group
+
+        :param group_id: id of group
+        :param client_id: id of client (not client-id)
+        :param roles: roles list or role (use GroupRoleRepresentation)
+        :return Keycloak server response
+        """
+
+        payload = roles if isinstance(roles, list) else [roles]
+        params_path = {"realm-name": self.realm_name, "id": group_id, "client-id": client_id}
+        data_raw = self.raw_post(URL_ADMIN_GROUPS_CLIENT_ROLES.format(**params_path),
+                                 data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+
+    def delete_group_client_roles(self, group_id, client_id, roles):
+        """
+        Delete client roles of a group
+
+        :param group_id: id of group
+        :param client_id: id of client (not client-id)
+        :param roles: roles list or role (use GroupRoleRepresentation)
+        :return Keycloak server response
+        """
+
+        payload = roles if isinstance(roles, list) else [roles]
+        params_path = {"realm-name": self.realm_name, "id": group_id, "client-id": client_id}
+        data_raw = self.raw_get(URL_ADMIN_GROUPS_CLIENT_ROLES.format(**params_path))
+        return raise_error_from_response(data_raw, KeycloakGetError)
+
+    def get_group_client_roles(self, group_id, client_id, roles):
+        """
+        Get client roles of a group
+
+        :param group_id: id of group
+        :param client_id: id of client (not client-id)
+        :param roles: roles list or role (use GroupRoleRepresentation)
+        :return Keycloak server response (array RoleRepresentation)
+        """
+
+        payload = roles if isinstance(roles, list) else [roles]
+        params_path = {"realm-name": self.realm_name, "id": group_id, "client-id": client_id}
+        data_raw = self.raw_delete(URL_ADMIN_GROUPS_CLIENT_ROLES.format(**params_path),
+                                 data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+
     def get_client_roles_of_user(self, user_id, client_id):
         """
         Get all client roles for a user.
@@ -1080,7 +1151,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "id": user_id, "client-id": client_id}
         data_raw = self.raw_delete(URL_ADMIN_USER_CLIENT_ROLES.format(**params_path),
                                    data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def get_authentication_flows(self):
         """
@@ -1110,7 +1181,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name}
         data_raw = self.raw_post(URL_ADMIN_FLOWS.format(**params_path),
                                  data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201], skip_exists=skip_exists)
 
     def get_authentication_flow_executions(self, flow_alias):
         """
@@ -1138,7 +1209,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "flow-alias": flow_alias}
         data_raw = self.raw_put(URL_ADMIN_FLOWS_EXECUTIONS.format(**params_path),
                                 data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def sync_users(self, storage_id, action):
         """
@@ -1196,7 +1267,7 @@ class KeycloakAdmin:
         data_raw = self.raw_post(
             URL_ADMIN_CLIENT_SCOPES_ADD_MAPPER.format(**params_path), data=json.dumps(payload))
 
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201])
 
     def generate_client_secrets(self, client_id):
         """
@@ -1256,7 +1327,7 @@ class KeycloakAdmin:
 
         data_raw = self.raw_post(URL_ADMIN_COMPONENTS.format(**params_path),
                                  data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[201])
 
     def get_component(self, component_id):
         """
@@ -1286,7 +1357,7 @@ class KeycloakAdmin:
         params_path = {"realm-name": self.realm_name, "component-id": component_id}
         data_raw = self.raw_put(URL_ADMIN_COMPONENT.format(**params_path),
                                 data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def delete_component(self, component_id):
         """
@@ -1298,7 +1369,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "component-id": component_id}
         data_raw = self.raw_delete(URL_ADMIN_COMPONENT.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
 
     def get_keys(self):
         """
@@ -1397,7 +1468,8 @@ class KeycloakAdmin:
         try:
             self.token = self.keycloak_openid.refresh_token(refresh_token)
         except KeycloakGetError as e:
-            if e.response_code == 400 and b'Refresh token expired' in e.response_body:
+            if e.response_code == 400 and (b'Refresh token expired' in e.response_body or
+                                           b'Token is not active' in e.response_body):
                 self.get_token()
             else:
                 raise
